/****************************************************************************
**
** Copyright (C) 2010 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    Larvor.cpp
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @brief   Interface between serial port and JTAG/SPI for programming AVR microcontrollers
// @web     http://sophiateam.undrgnd.free.fr
// @date    08/08/2010
// ---------------------------------------------------------------------------

#include "Larvor.h"
#include "sirSerial.h"
#include "larvorProtocol.h"
#include "crc.h"

#include <QString>

// ----------------------------------------------------------------------------
// Constructor
// ----------------------------------------------------------------------------
Larvor::Larvor(sirSerial* port)
    : serialPort_(port),
    programmingMode_(jtagProgrammingMode)
{
}


// ----------------------------------------------------------------------------
// ProgrammingMode
// ----------------------------------------------------------------------------
void
Larvor::setProgrammingMode(Larvor::ProgrammingMode m)
{
    programmingMode_ = m;
}

Larvor::ProgrammingMode Larvor::getProgrammingMode() const
{
    return programmingMode_;
}

// ----------------------------------------------------------------------------
// Test the serial port connection between the PC and Larvor. 
// Return true on success 
// ----------------------------------------------------------------------------
bool 
Larvor::ping()
{
    //reset(true);
    write(PROG_PING);
    return (read() == PROG_PING);
}

// ----------------------------------------------------------------------------
// leds
// ----------------------------------------------------------------------------
void 
Larvor::ledOff(int ledId)
{
    write(ledId==0?PROG_SET_LED_1:PROG_SET_LED_2);
    write(0x00); // blink slow
}

void 
Larvor::ledBlinkSlow(int ledId)
{
    write(ledId==0?PROG_SET_LED_1:PROG_SET_LED_2);
    write(0x03); // blink slow
}

void 
Larvor::ledBlinkFast(int ledId)
{
    write(ledId==0?PROG_SET_LED_1:PROG_SET_LED_2);
    write(0x55); // blink slow
}

void 
Larvor::ledOn(int ledId)
{
    write(ledId==0?PROG_SET_LED_1:PROG_SET_LED_2);
    write(0xFF); // blink slow
}


LarvorLedBlinker::LarvorLedBlinker(Larvor * larvor, int ledId)
    : larvor_(larvor), ledId_(ledId)
{
    if (larvor_)
        larvor_->ledBlinkFast(ledId_);
}

LarvorLedBlinker::~LarvorLedBlinker()
{
    if (larvor_)
        larvor_->ledOff(ledId_);
}

// ----------------------------------------------------------------------------
// update the progress bar and return false if the progress dialog was canceled
// (user want to stop the current operation)
// ----------------------------------------------------------------------------
bool Larvor::setProgress(double v)
{
    progress_.setValue(v);
    return !progress_.canceled();
}
// ----------------------------------------------------------------------------
// Read one byte on the serial port
// ----------------------------------------------------------------------------
//#define LARVOR_SERIAL_DUMP

static const int LARVOR_SERIAL_RETRY = 20; // retry until data
unsigned char 
Larvor::read(bool* ok)
{
    unsigned char data;
    unsigned char buf[2];
    unsigned long length=1;
    if (serialPort_ && serialPort_->forceRead(buf, length, LARVOR_SERIAL_RETRY)) {
        data = buf[0];
        if (ok) *ok= true;
    } else {
        data = 0;
        if (ok) *ok= false;
    }
    return data;
}

unsigned short 
Larvor::read16(bool* ok)
{
    bool ok1, ok2;
    unsigned short data = read(&ok1) + (read(&ok2)<<8);
    if (ok) {
        *ok = ok1 && ok2;
    }
    return data;
}

unsigned long 
Larvor::read32(bool* ok)
{
    bool ok1, ok2, ok3, ok4;
    unsigned long data = read(&ok1) + (read(&ok2)<<8) + (read(&ok3)<<16) + (read(&ok4)<<24);
    if (ok) {
        *ok = ok1 && ok2 && ok3 && ok4;
    }
    return data;
}

// ----------------------------------------------------------------------------
// write one byte on the serial port
// ----------------------------------------------------------------------------
void
Larvor::write(unsigned char data, 
              bool* ok)
{
    if (serialPort_) {
        serialPort_->write(data);
        if (ok) *ok= true;
    } else {
        data = 0;
        if (ok) *ok= false;
    }
}
// ----------------------------------------------------------------------------
// write one byte on the serial port
// ----------------------------------------------------------------------------
void
Larvor::write16(unsigned short data, 
                bool* ok)
{
    if (serialPort_) {
        serialPort_->write(data & 0xFF);
        serialPort_->write((data>>8) & 0xFF);
        if (ok) *ok= true;
    } else {
        data = 0;
        if (ok) *ok= false;
    }
}

// ----------------------------------------------------------------------------
//
// ----------------------------------------------------------------------------
bool 
Larvor::writeAddrWidthCrc(unsigned long addrBegin,
                          unsigned long addrEnd,
                          unsigned char pageSize)
{
    unsigned char crc=0;
    crc = CRC16(addrBegin, crc);
    crc = CRC16(addrEnd, crc);
    crc = CRC8(pageSize, crc);
    write16(addrBegin);
    write16(addrEnd);
    write(pageSize);
    write(crc);
    return (read() == PROG_CRC_OK);
}

// ----------------------------------------------------------------------------
// Send a synchronisation sequence after a CRC error
// ----------------------------------------------------------------------------
void
Larvor::syncCom()
{
    write(PROG_SYNC1);
    write(PROG_SYNC1);
    write(PROG_SYNC1);
    write(PROG_SYNC2);
}
