/****************************************************************************
**
** Copyright (C) 2010 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    Larvor.h
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @brief   Interface between serial port and JTAG/SPI for programming AVR microcontrollers
// @web     http://sophiateam.undrgnd.free.fr
// @date    08/08/2010
// ---------------------------------------------------------------------------

#ifndef __LARVOR_H__
#define __LARVOR_H__

#include <stdio.h>
#include <QMap>
#include <QString>
#include <QWidget>
#include "sirSerial.h"
class sirSerial;
class Larvor;

// ----------------------------------------------------------------------------
// LarvorFiles : read write eeprom and flash files generated by avrgcc
// ----------------------------------------------------------------------------
namespace LarvorFiles
{

    bool readFlashFile(QString const& filename, 
                       QMap<unsigned long, unsigned short> & flashBuf) ;
    bool writeFlashIntelHex (QString const& filename, 
                             QMap<unsigned long, unsigned short> const& buffer);
    
    bool readEepromFile(QString const& filename, 
                        QMap<unsigned long, unsigned char> & eepromBuf);
    bool writeEepromIntelHex (QString const& filename, 
                              QMap<unsigned long, unsigned char> const& buffer); 

    void adjustToPageSize(QMap<unsigned long, unsigned char>& data,
                          unsigned int pagesize,
                          unsigned long &addrBegin,
                          unsigned long &addrEnd);

    void adjustToPageSize(QMap<unsigned long, unsigned short>& data,
                          unsigned int pagesize,
                          unsigned long &addrBegin,
                          unsigned long &addrEnd);

    bool checkFlash(QMap<unsigned long, unsigned short> const& data,
                    QMap<unsigned long, unsigned short> const& data2);

    bool checkEEprom(QMap<unsigned long, unsigned char> const& data,
                     QMap<unsigned long, unsigned char> const& data2);
    
}

// ----------------------------------------------------------------------------
// LarvorDevices: contains info about the devices supported by Larvor
// ----------------------------------------------------------------------------
namespace LarvorDevices
{
    // Device informations
    struct DeviceInfo {
        int           deviceId;
        QString       deviceName;
        int           signature;
        int           manufacturerId;
        QString       manufacturerName;
        unsigned long eepromSize;
        int           eepromPageSize;
        unsigned long flashSize;
        int           flashPageSize;

        QMap<int, QString> lowFuseDescription;
        QMap<int, QString> highFuseDescription;
        QMap<int, QString> extendedFuseDescription;

        DeviceInfo();

        bool isUnknown() { return deviceId < 0; }
    };

    // load an XML file containing the definition of the devices
    int loadDeviceConfigFile(QString const& filename);

    // return the device info associated to a deviceId
    bool findDeviceInfo(int manufacturerId, int deviceId,
                        DeviceInfo &info);
    bool findDeviceInfo(unsigned char signature1, unsigned char signature2, unsigned char signature3,
                        DeviceInfo & info);

    // get the fuse description
    QString fuseHighToString(DeviceInfo const& info,
                             unsigned char fuse);
    QString fuseLowToString(DeviceInfo const& info,
                            unsigned char fuse);
    QString fuseExtendedToString(DeviceInfo const& info,
                                 unsigned char fuse);
    
    QString lockToString(unsigned char lock);
}

// ----------------------------------------------------------------------------
// LarvorProgress
// ----------------------------------------------------------------------------
class LarvorProgress : public QWidget {
    Q_OBJECT

public:
    LarvorProgress() : QWidget(), value_(0), canceled_(false) {}
    void   setValue(double v) {
        value_ = v;
        valueChanged(v);
    }
    void   setCanceled(bool c) { canceled_ = c; }

    double value() const { return value_; }
    bool   canceled() const { return canceled_; }

signals:
    void valueChanged(double);

private:
    double value_;
    bool   canceled_;
};

// ----------------------------------------------------------------------------
// LarvorLedBlinker
// ----------------------------------------------------------------------------

class LarvorLedBlinker
{
public:
    LarvorLedBlinker(Larvor * larvor, int ledId);
    ~LarvorLedBlinker();
private:
    Larvor *larvor_;
    int    ledId_;
};

// ----------------------------------------------------------------------------
// Larvor
// ----------------------------------------------------------------------------
// Abstract class for programming a device with the larvor board
// ----------------------------------------------------------------------------
class Larvor {
public:
    static const int ledRed=0;
    static const int ledBlue=1;

    enum ProgrammingMode {
        spiProgrammingMode=0,
        jtagProgrammingMode=1
    };

public:
    Larvor(sirSerial* port);
    virtual ~Larvor() {}

    void setProgrammingMode(ProgrammingMode m);
    ProgrammingMode getProgrammingMode() const;

    // ping Larvor board. Return true if Larvor board is connected to the serial port
    bool ping();
    // leds
    void ledOff(int ledId);
    void ledBlinkSlow(int ledId);
    void ledBlinkFast(int ledId);
    void ledOn(int ledId);

    // programmation
    virtual bool chipErase();

    virtual bool getDeviceId(int& manufacturerId,  int& deviceId);
    virtual bool readSignature(unsigned char& signature1,
                               unsigned char& signature2,
                               unsigned char& signature3);

    virtual bool readCalibration(unsigned char& calibration);

    virtual bool readFuseExt(unsigned char& fuseExt);
    virtual bool readFuseHigh(unsigned char& fuseHigh);
    virtual bool readFuseLow( unsigned char& fuseLow);
    virtual bool writeFuseExt(unsigned char fuseExt);
    virtual bool writeFuseHigh(unsigned char fuseHigh);
    virtual bool writeFuseLow(unsigned char fuseHigh);

    virtual bool readLock(unsigned char& lock);
    virtual bool writeLock(unsigned char lock);

    virtual bool readEEprom(unsigned long addrBegin,
                            unsigned long addrEnd,
                            QMap<unsigned long, unsigned char>& data);
    
    virtual bool writeEEprom(unsigned long addrBegin,
                             unsigned long addrEnd,
                             QMap<unsigned long, unsigned char> const& data,
                             unsigned char pageSize);

    virtual bool readFlash(unsigned long addrBegin,
                           unsigned long addrEnd,
                           QMap<unsigned long, unsigned short>& data);
    virtual bool writeFlash(unsigned long addrBegin,
                            unsigned long addrEnd,
                            QMap<unsigned long, unsigned short> const& data,
                            unsigned char pageSize);

    // return the progressBar
    LarvorProgress* getProgress() { return &progress_; }

    
    
protected:
    void usleep(unsigned long t) { if (serialPort_) serialPort_->usleep(t); }
    // read one byte from serial port
    unsigned char  read(bool* ok=NULL);
    unsigned short read16(bool* ok=NULL);
    unsigned long  read32(bool* ok=NULL);
    // write one jtag/spi command on the serial port depending on the programmingMode_
    void           writeCmd(unsigned char cmd);
    // write one byte on serial port
    void           write(unsigned char data,
                         bool* ok=NULL);
    void           write16(unsigned short data,
                           bool* ok=NULL);
    bool           writeAddrWidthCrc(unsigned long addrBegin,
                                     unsigned long addrEnd,
                                     unsigned char pageSize);
    // send a synchronisation sequence after a CRC error
    void           syncCom();
    bool           setProgress(double v);

private:
    sirSerial*      serialPort_;
    LarvorProgress  progress_;
    ProgrammingMode programmingMode_;
};



#endif
