/****************************************************************************
**
** Copyright (C) 2010 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    LarvorCmd.cpp
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @web     http://sophiateam.undrgnd.free.fr
// @date    08/08/2010
// ---------------------------------------------------------------------------


#include "Larvor.h"
#include "larvorProtocol.h"
#include "crc.h"

#include <QString>


#define PROG_GET_DEVICE_ID          0x20   // IN: - OUT: [4 bytes signature] [PROG_OK]
#define PROG_READ_SIGNATURE         0x21   // IN: - OUT: [signa1] [signa2] [signa3][0x00] [PROG_OK]
#define PROG_READ_CALIBRATION       0x22   // IN: - OUT: [calibration] [PROG_OK]
#define PROG_READ_FUSE_EXT          0x23   // IN: - OUT: [fuse_ext] [PROG_OK]
#define PROG_READ_FUSE_HIGH         0x24   // IN: - OUT: [fuse_high] [PROG_OK]
#define PROG_READ_FUSE_LOW          0x25   // IN: - OUT: [fuse_low] [PROG_OK]
#define PROG_READ_LOCK              0x26   // IN: - OUT: [lock] [PROG_OK]
#define PROG_WRITE_FUSE_EXT         0x27   // IN: [fuse_ext][CRC8]  - OUT: [PROG_OK] / [PROG_ERROR]
#define PROG_WRITE_FUSE_HIGH        0x28   // IN: [fuse_high][CRC8] - OUT: [PROG_OK] / [PROG_ERROR]
#define PROG_WRITE_FUSE_LOW         0x29   // IN: [fuse_low][CRC8]  - OUT: [PROG_OK] / [PROG_ERROR]
#define PROG_WRITE_LOCK             0x2A   // IN: [lock][CRC8]      - OUT: [PROG_OK] / [PROG_ERROR]
#define PROG_CHIP_ERASE             0x2B   // IN: - OUT: [PROG_OK]
#define PROG_READ_EEPROM            0x2C   // IN: [addrBeginL] [addrBeginH] [addrEndL] [addrEndH] [CRC8] - OUT: [one byte per data] [...] [PROG_OK]
#define PROG_WRITE_EEPROM           0x2D   // IN: [addrBeginL] [addrBeginH] [addrEndL] [addrEndH] [CRC8] [one byte per data] [...] - OUT: [PROG_OK]
#define PROG_READ_FLASH             0x2E   // IN: [addrBeginL] [addrBeginH] [addrEndL] [addrEndH] [CRC8] - OUT: [dataL][dataH] [...] [PROG_OK]
#define PROG_WRITE_FLASH            0x2F

void
Larvor::writeCmd(unsigned char cmd)
{
    unsigned char larvorCmd = cmd;

    if (programmingMode_ == spiProgrammingMode) {
        switch(cmd) {
        case PROG_GET_DEVICE_ID:
            larvorCmd = PROG_SPI_GET_DEVICE_ID;
            break;
        case PROG_READ_SIGNATURE:
            larvorCmd = PROG_SPI_READ_SIGNATURE;
            break;
        case PROG_READ_CALIBRATION:
            larvorCmd = PROG_SPI_READ_CALIBRATION;
            break;
        case PROG_READ_FUSE_EXT:
            larvorCmd = PROG_SPI_READ_FUSE_EXT;
            break;
        case PROG_READ_FUSE_HIGH:
            larvorCmd = PROG_SPI_READ_FUSE_HIGH;
            break;
        case PROG_READ_FUSE_LOW:
            larvorCmd = PROG_SPI_READ_FUSE_LOW;
            break;
        case PROG_READ_LOCK:
            larvorCmd = PROG_SPI_READ_LOCK;
            break;
        case PROG_WRITE_FUSE_EXT:
            larvorCmd = PROG_SPI_WRITE_FUSE_EXT;
            break;
        case PROG_WRITE_FUSE_HIGH:
            larvorCmd = PROG_SPI_WRITE_FUSE_HIGH;
            break;
        case PROG_WRITE_FUSE_LOW:
            larvorCmd = PROG_SPI_WRITE_FUSE_LOW;
            break;
        case PROG_CHIP_ERASE:
            larvorCmd = PROG_SPI_CHIP_ERASE;
            break;
        case PROG_READ_EEPROM:
            larvorCmd = PROG_SPI_READ_EEPROM;
            break;
        case PROG_WRITE_EEPROM:
            larvorCmd = PROG_SPI_WRITE_EEPROM;
            break;
        case PROG_READ_FLASH:
            larvorCmd = PROG_SPI_READ_FLASH;
            break;
        case PROG_WRITE_FLASH:
            larvorCmd = PROG_SPI_WRITE_FLASH;
            break;
        }
    } else if (programmingMode_ == jtagProgrammingMode) {
        switch(cmd) {
        case PROG_GET_DEVICE_ID:
            larvorCmd = PROG_JTAG_GET_DEVICE_ID;
            break;
        case PROG_READ_SIGNATURE:
            larvorCmd = PROG_JTAG_READ_SIGNATURE;
            break;
        case PROG_READ_CALIBRATION:
            larvorCmd = PROG_JTAG_READ_CALIBRATION;
            break;
        case PROG_READ_FUSE_EXT:
            larvorCmd = PROG_JTAG_READ_FUSE_EXT;
            break;
        case PROG_READ_FUSE_HIGH:
            larvorCmd = PROG_JTAG_READ_FUSE_HIGH;
            break;
        case PROG_READ_FUSE_LOW:
            larvorCmd = PROG_JTAG_READ_FUSE_LOW;
            break;
        case PROG_READ_LOCK:
            larvorCmd = PROG_JTAG_READ_LOCK;
            break;
        case PROG_WRITE_FUSE_EXT:
            larvorCmd = PROG_JTAG_WRITE_FUSE_EXT;
            break;
        case PROG_WRITE_FUSE_HIGH:
            larvorCmd = PROG_JTAG_WRITE_FUSE_HIGH;
            break;
        case PROG_WRITE_FUSE_LOW:
            larvorCmd = PROG_JTAG_WRITE_FUSE_LOW;
            break;
        case PROG_CHIP_ERASE:
            larvorCmd = PROG_JTAG_CHIP_ERASE;
            break;
        case PROG_READ_EEPROM:
            larvorCmd = PROG_JTAG_READ_EEPROM;
            break;
        case PROG_WRITE_EEPROM:
            larvorCmd = PROG_JTAG_WRITE_EEPROM;
            break;
        case PROG_READ_FLASH:
            larvorCmd = PROG_JTAG_READ_FLASH;
            break;
        case PROG_WRITE_FLASH:
            larvorCmd = PROG_JTAG_WRITE_FLASH;
            break;
        }
    }
    write(larvorCmd);
}

// ============================================================================
// AVR Jtag commands
// ============================================================================


// ----------------------------------------------------------------------------
// Get the device ID code from the DR register after reseting the jtag
// ----------------------------------------------------------------------------
bool
Larvor::getDeviceId(int&     manufacturerId,
                    int&     deviceId)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_GET_DEVICE_ID);

    // read 32 bits
    uint32_t data = read32();

    bool valid     = ((data & 0x1) != 0); // first bit must be 1
    manufacturerId = (data >> 1)  & 0x7FF;  // 11 bits
    deviceId       = (data >> 12) & 0xFFFF; // 16 bits

    return valid && (read() == PROG_OK);
}

// ----------------------------------------------------------------------------
// AVR Reading the signature bytes
// ----------------------------------------------------------------------------
bool
Larvor::readSignature(unsigned char& signature1,
                          unsigned char& signature2,
                          unsigned char& signature3)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_SIGNATURE);
    uint32_t data = read32()&0xFFFFFF;
    signature1 = data & 0xFF;
    signature2 = (data >> 8) & 0xFF;
    signature3 = (data >> 16) & 0xFF;

    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the Calibration byte
// ----------------------------------------------------------------------------
bool 
Larvor::readCalibration(unsigned char& calibration)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_SIGNATURE);
    calibration = read();
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the Fuse High byte
// ----------------------------------------------------------------------------
bool 
Larvor::readFuseHigh(unsigned char& fuseHigh)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_FUSE_HIGH);
    fuseHigh = read();
    return read() == PROG_OK;
}
// ----------------------------------------------------------------------------
// AVR Reading the Fuse Low byte
// ----------------------------------------------------------------------------
bool 
Larvor::readFuseLow( unsigned char& fuseLow)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_FUSE_LOW);
    fuseLow = read();
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the Fuse Low byte
// ----------------------------------------------------------------------------
bool 
Larvor::readFuseExt( unsigned char& fuseExt)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_FUSE_EXT);
    fuseExt = read();
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the Lock byte
// ----------------------------------------------------------------------------
bool 
Larvor::readLock(unsigned char& lock)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_LOCK);
    lock = read();
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Writing the Fuse High byte
// ----------------------------------------------------------------------------
bool 
Larvor::writeFuseHigh(unsigned char fuseHigh)
{
    LarvorLedBlinker blinker(this, ledBlue);
    unsigned char crc = CRC8(fuseHigh, 0);
    writeCmd(PROG_WRITE_FUSE_HIGH);
    write(fuseHigh);
    write(crc);
    return read() == PROG_OK;
}
// ----------------------------------------------------------------------------
// AVR Writing the Fuse Low bytes
// ----------------------------------------------------------------------------
bool 
Larvor::writeFuseLow(unsigned char fuseLow)
{
    LarvorLedBlinker blinker(this, ledBlue);
    unsigned char crc = CRC8(fuseLow, 0);
    writeCmd(PROG_WRITE_FUSE_LOW);
    write(fuseLow);
    write(crc);
    return read() == PROG_OK;
}
// ----------------------------------------------------------------------------
// AVR Writing the Fuse Extended bytes
// ----------------------------------------------------------------------------
bool 
Larvor::writeFuseExt(unsigned char fuseExt)
{
    LarvorLedBlinker blinker(this, ledBlue);
    unsigned char crc = CRC8(fuseExt, 0);
    writeCmd(PROG_WRITE_FUSE_EXT);
    write(fuseExt);
    write(crc);
    return read() == PROG_OK;
}
// ----------------------------------------------------------------------------
// AVR Writing the Lock bytes
// ----------------------------------------------------------------------------
bool 
Larvor::writeLock(unsigned char lock)
{
    LarvorLedBlinker blinker(this, ledBlue);
    unsigned char crc = CRC8(lock, 0);
    writeCmd(PROG_WRITE_LOCK);
    write(lock);
    write(crc);
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the eeprom from addrBegin to addrEnd (addrEnd is included)
// ----------------------------------------------------------------------------
bool 
Larvor::readEEprom(unsigned long addrBegin,
                   unsigned long addrEnd,
                   QMap<unsigned long, unsigned char>& data)
{
    bool status = true;
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_EEPROM);
    if (!writeAddrWidthCrc(addrBegin, addrEnd, 0))
        return false;


    unsigned char crc=0, i=0;
    int crcRetry = 5;
    for (unsigned long addr = addrBegin; addr <= addrEnd; ++addr) {
        data[addr] = read();

        // CRC
        crc = CRC8(data[addr], crc);
        ++i;
        if (i == PROG_CRC_CHECK_SIZE) { // compare CRC with the target
            i = 0;
            unsigned char crcExpected = read();
            if (getProgress()->canceled()) {
                write(PROG_CANCEL);
                break;
            } else if (crcExpected != crc && --crcRetry > 0) {
                // read error: read the same block again
                addr -= PROG_CRC_CHECK_SIZE;
                write(PROG_CRC_ERROR);
            } else {
                // read OK, move to next block
                status &= (crcExpected == crc);
                crcRetry = 5;
                write(PROG_CRC_OK);
            }
            crc = 0;
        }

        // update the progressBar
        double p = (double)(addr-addrBegin)/(double)(addrEnd-addrBegin);
        setProgress(p);
    }

    status &= (read() == PROG_OK);

    return status;
}

// ----------------------------------------------------------------------------
// AVR Writing the eeprom
// ----------------------------------------------------------------------------
bool 
Larvor::writeEEprom(unsigned long addrBegin,
                    unsigned long addrEnd,
                    QMap<unsigned long, unsigned char> const & data,
                    unsigned char pageSize)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_WRITE_EEPROM);
    if (!writeAddrWidthCrc(addrBegin, addrEnd, pageSize))
        return false;

    unsigned long addr;
    for(addr=addrBegin; addr <= addrEnd;) {
        int retry = 0;
        while(retry++ < 10) {
            // send status byte
            write(PROG_OK);
            // send the data and compute the CRC
            int indexInPage;
            unsigned char crc=0;
            for(indexInPage = 0; indexInPage < pageSize; ++indexInPage, ++addr) {
                unsigned char  v = data[addr] & 0xFF;
                write(v);
                // CRC
                crc = CRC8(v, crc);
            }
            // send the CRC
            write(crc);
            if (read() != PROG_CRC_OK) {
                syncCom();
                addr -= pageSize;
            } else {
                break;
            }
        }
        if (retry >= 10) {
            write(PROG_CANCEL);
            break;
        }

        // update the progressBar
        double p = (double)(addr-addrBegin)/(double)(addrEnd-addrBegin);
        if (!setProgress(p)) {
            write(PROG_CANCEL);
            break;
        }
    }
    return read() == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Reading the flash
// ----------------------------------------------------------------------------
bool 
Larvor::readFlash(unsigned long addrBegin,
                  unsigned long addrEnd,
                  QMap<unsigned long, unsigned short>& data)
{
    bool status = true;
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_READ_FLASH);
    if (!writeAddrWidthCrc(addrBegin, addrEnd, 0))
        return false;
    
    unsigned char crc=0, i=0;
    int crcRetry = 5;
    for (unsigned long addr = addrBegin; addr <= addrEnd; ++addr) {
        unsigned short v = read16();
        data[addr] = v;

        // CRC
        crc = CRC16(data[addr], crc);
        ++i;
        if (i == PROG_CRC_CHECK_SIZE) { // compare CRC with the target
            i = 0;
            unsigned char crcExpected = read();
            if (getProgress()->canceled()) {
                write(PROG_CANCEL);
                break;
            } else if (crcExpected != crc && --crcRetry > 0) {
                addr -= PROG_CRC_CHECK_SIZE;
                write(PROG_CRC_ERROR);
            } else {
                status &= (crcExpected == crc);
                crcRetry = 5;
                write(PROG_CRC_OK);
            }
            crc = 0;
        }

        // update the progressBar
        double p = (double)(addr-addrBegin)/(double)(addrEnd-addrBegin);
        setProgress(p);
    } 
    status &= (read() == PROG_OK);
    return status;
}
// ----------------------------------------------------------------------------
// AVR Writing the flash
// ----------------------------------------------------------------------------
bool 
Larvor::writeFlash(unsigned long addrBegin,
                   unsigned long addrEnd,
                   QMap<unsigned long, unsigned short> const& data,
                   unsigned char pageSize)
{
    LarvorLedBlinker blinker(this, ledBlue);

    writeCmd(PROG_WRITE_FLASH);
    if (!writeAddrWidthCrc(addrBegin, addrEnd, pageSize))
        return false;
    unsigned long addr;
    for(addr=addrBegin; addr <= addrEnd;) {
        int retry = 0;
        while(retry++ < 10) {
            // send status byte
            write(PROG_OK);
            // send the data and compute the CRC
            unsigned char indexInPage;
            unsigned char crc=0;
            for(indexInPage = 0; indexInPage < pageSize; ++indexInPage, ++addr) {
                unsigned short v    = data[addr];
                write16(v);
                // CRC
                crc = CRC16(v, crc);
                //usleep(100);
            }
            // send the CRC
            write(crc);
            // check that data have been received correctly
            unsigned char r=0;
            do {
                r = read();
            } while(r == 0);
            if (r != PROG_CRC_OK) {
                syncCom();
                addr -= pageSize;
            } else {
                break;
            }

        }
        if (retry >= 10) {
            write(PROG_CANCEL);
            break;
        }

        // update the progressBar
        double p = (double)(addr-addrBegin)/(double)(addrEnd-addrBegin);
        if (!setProgress(p)) {
            write(PROG_CANCEL);
            break;
        }
    }
    unsigned char r=0;
    do {
        r = read();
    } while(r == 0);
    return r == PROG_OK;
}

// ----------------------------------------------------------------------------
// AVR Erasing the chip
// ----------------------------------------------------------------------------
bool 
Larvor::chipErase()
{
    LarvorLedBlinker blinker(this, ledBlue);
    writeCmd(PROG_CHIP_ERASE);
    usleep(20000);
    return read() == PROG_OK;
}
