/****************************************************************************
**
** Copyright (C) 2010 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    LarvorDevices.cpp
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @web     http://sophiateam.undrgnd.free.fr
// @date    08/08/2010
// ---------------------------------------------------------------------------

#include "larvor.h"
#include <QFile>
#include <Qmap>
#include <QXmlStreamReader> 
#include <QMessageBox>

// private functions of the LarvorDevices namespace
namespace LarvorDevices
{
    QMap<int/*deviceId*/, DeviceInfo> devices_;
    void registerDevice(DeviceInfo const& info);
	QString fuseToString(unsigned char             fuse,
						 QString const&            title,
						 QMap<int, QString> const& descriptions);
}

// ---------------------------------------------------------------------------
// struct DeviceInfo
// ---------------------------------------------------------------------------

LarvorDevices::DeviceInfo::DeviceInfo() :
        deviceId(-1),
        deviceName("Unknown device"),
		signature(0),
        manufacturerId(-1),
        manufacturerName("Unknown manufacturer"),
        eepromSize(0),
        eepromPageSize(1),
        flashSize(0),
        flashPageSize(1)
{
}

// ---------------------------------------------------------------------------
// class LarvorDeviceXmlReader
// ---------------------------------------------------------------------------
class LarvorDeviceXmlReader : public QXmlStreamReader
{
public:
    bool read(QString const& filename);
    
private:
    void readUnknownElement();
    void readLarvorDevicesFile();
    void readDeviceManufacturer();
    void readDevice(int manufacturerId, QString const& manufacturerName);
    void readDeviceEEprom(LarvorDevices::DeviceInfo& device);
    void readDeviceFlash(LarvorDevices::DeviceInfo& device);
    void readDeviceFuse(QMap<int, QString> & fuseDescription);
};

bool LarvorDeviceXmlReader::read(QString const& filename)
{
    QFile file(filename);
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
        QMessageBox::warning(NULL, "Larvor devices file",
                             QString("Cannot read file %1:\n%2.")
                             .arg(filename)
                             .arg(file.errorString()));
        return false;
    }
    setDevice(&file);

    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            if (name() == "LarvorDevicesFile" && attributes().value("version") == "1.0") {
                readLarvorDevicesFile();
            } else {
                readUnknownElement();
            }
        }
        if (isEndElement()) {
            break;
        }
    }
    return true;
}

void LarvorDeviceXmlReader::readLarvorDevicesFile()
{
    QString elemName = name().toString();
    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            if (name() == "manufacturer")
                readDeviceManufacturer();
            else {
                readUnknownElement();
            }
        }
        if (isEndElement() && name() == elemName) {
            break;
        }
    }
}
void LarvorDeviceXmlReader::readDeviceManufacturer()
{
    Q_ASSERT(isStartElement() && name() == "manufacturer");
    int manufacturerId       = attributes().value("id").toString().toInt(NULL, 16);
    QString manufacturerName = attributes().value("name").toString ();

    QString elemName = name().toString();
    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            if (name() == "device")
                readDevice(manufacturerId, manufacturerName);
            else {
                readUnknownElement();
            }
        }
        if (isEndElement() && name() == elemName) {
            break;
        }
    }
}

void LarvorDeviceXmlReader::readDevice(int manufacturerId,
                                       QString const& manufacturerName)
{
    Q_ASSERT(isStartElement() && name() == "device");
    LarvorDevices::DeviceInfo device;
    bool ok;
    device.deviceId   = attributes().value("id").toString().toInt(&ok, 16);
    if (!ok) device.deviceId = -1;
    device.signature  = attributes().value("signature").toString().toInt(&ok, 16);
    if (!ok) device.signature = 0;
    device.deviceName = attributes().value("name").toString();
    device.manufacturerId   = manufacturerId;
    device.manufacturerName = manufacturerName;

    QString elemName = name().toString();
    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            if (name() == "eeprom")
                readDeviceEEprom(device);
            else if (name() == "flash")
                readDeviceFlash(device);
            else if (name() == "lowFuse")
                readDeviceFuse(device.lowFuseDescription);
            else if (name() == "highFuse")
                readDeviceFuse(device.highFuseDescription);
            else if (name() == "extendedFuse")
                readDeviceFuse(device.extendedFuseDescription);
            else {
                readUnknownElement();
            }
        }
        if (isEndElement() && (name() == elemName)) {
            break;
        }
    }

    if (!device.isUnknown()) {
        LarvorDevices::registerDevice(device);
    }
}
void LarvorDeviceXmlReader::readDeviceEEprom(LarvorDevices::DeviceInfo& device)
{
    bool ok;
    device.eepromSize       = attributes().value("size").toString().toInt(&ok, 16);
    if (!ok) device.eepromSize = 0;
    device.eepromPageSize   = attributes().value("pageSize").toString().toInt(&ok);
    if (!ok) device.eepromPageSize = 1;

    QString elemName = name().toString();
    while (!atEnd() && !(isEndElement() && name() == elemName)) readNext();
}

void LarvorDeviceXmlReader::readDeviceFlash(LarvorDevices::DeviceInfo& device)
{
    bool ok;
    device.flashSize        = attributes().value("size").toString().toInt(&ok, 16);
    if (!ok) device.flashSize = 0;
    device.flashPageSize    = attributes().value("pageSize").toString().toInt(&ok);
    if (!ok) device.flashPageSize = 1;

    QString elemName = name().toString();
    while (!atEnd() && !(isEndElement() && name() == elemName)) readNext();
}

void LarvorDeviceXmlReader::readDeviceFuse(QMap<int, QString> & fuseDescription)
{
    QString elemName = name().toString();
    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            if (name() == "fuse") {
                bool ok;
                int bit = attributes().value("bit").toString().toInt(&ok);
                if (!ok) bit = 0;
                int length = attributes().value("length").toString().toInt(&ok);
                if (!ok) length = 1;
                if (!isEndElement() && name() == "fuse") readNext();
                if (isCharacters ())
                    fuseDescription[(bit << 8) + length] = text().toString();
                while (!atEnd() && !(isEndElement() && name() == "fuse")) readNext();
            } else {
                readUnknownElement();
            }
        }
        if (isEndElement() && (name() == elemName)) {
            break;
        }
    }

    while (!atEnd() && !(isEndElement() && name() == elemName)) readNext();
}

void LarvorDeviceXmlReader::readUnknownElement()
{
    QString elemName = name().toString();
    while (!atEnd()) {
        readNext();
        if (isStartElement()) {
            readUnknownElement();
        }
        if (isEndElement() && (name() == elemName))
            break;
    }
}

// ---------------------------------------------------------------------------
// LarvorDevices::loadDeviceConfigFile
// ---------------------------------------------------------------------------
int
LarvorDevices::loadDeviceConfigFile(QString const& filename)
{
    LarvorDeviceXmlReader xmlParser;
    xmlParser.read(filename);
    return devices_.size();
}



// ---------------------------------------------------------------------------
// LarvorDevices::findDeviceInfo
// ---------------------------------------------------------------------------
bool
LarvorDevices::findDeviceInfo(int manufacturerId,
                              int deviceId,
                              DeviceInfo &info)
{
    info.deviceId = -1;
    info.manufacturerId = -1;

    int key = deviceId + (manufacturerId<<16);

    if (devices_.find(key) != devices_.end()) {
        info = devices_.value(key);
    }
    return info.isUnknown() >= 0;
}

// ---------------------------------------------------------------------------
// LarvorDevices::findDeviceInfo
// ---------------------------------------------------------------------------
bool
LarvorDevices::findDeviceInfo(unsigned char signature0,
                              unsigned char signature1,
                              unsigned char signature2,
                              DeviceInfo & info)
{
    info.deviceId = -1;
    info.manufacturerId = -1;
    
    int signature = (signature0 << 16) + (signature1 << 8) + (signature2 << 0);

    QMapIterator<int, DeviceInfo> i(devices_);
    while (i.hasNext()) {
        i.next();
        if (i.value().signature == signature) {
            info = i.value();
            return true;
        }
    }
    return false;
}

// ---------------------------------------------------------------------------
// LarvorDevices::registerDevice
// ---------------------------------------------------------------------------
void
LarvorDevices::registerDevice(DeviceInfo const& info)
{
    int key = info.deviceId + (info.manufacturerId<<16);;
    devices_[key] = info;
}

// ----------------------------------------------------------------------------
// Convert the fuse high bits into readable string
// ----------------------------------------------------------------------------
QString
LarvorDevices::fuseToString(unsigned char             fuse,
                            QString const&            title,
                            QMap<int, QString> const& descriptions)
{
    QString txt ="";
    txt += "    " + title + ": 0x" + QString::number(fuse, 16) + " b" + QString::number(fuse, 2) + "\n";
    QMapIterator<int, QString> i(descriptions);
    while (i.hasNext()) {
        i.next();
        QString line;
        if (i.key() < 0) {
            line = i.value();
        } else {
            int bit = i.key() >> 8;
            int length = (i.key() & 0xFF);
            int v = (fuse >> bit) & ((0xFF) >> (8-length));
            if (length <= 1) {
                line = "      Bit " + QString::number(bit);
            } else {
                line = "      Bits " + QString::number(bit) + " to " + QString::number(bit+length-1);
            }
            QString txtBits= QString::number(v, 2);
            txtBits = txtBits.rightJustified (length, '0');
            line += " = " +  txtBits;
            line = line.leftJustified(25, ' ');
            line += i.value();
        }
        txt += line + "\n";
    }
    return txt;
}
// ----------------------------------------------------------------------------
// Convert the fuse high bits into readable string
// ----------------------------------------------------------------------------
QString
LarvorDevices::fuseHighToString(DeviceInfo const& info,
                                        unsigned char fuse)
{
    return fuseToString(fuse, "highFuse", info.highFuseDescription);
}
// ----------------------------------------------------------------------------
// Convert the fuse low bits into readable string
// ----------------------------------------------------------------------------
QString
LarvorDevices::fuseLowToString(DeviceInfo const& info,
                                       unsigned char fuse)
{
    return fuseToString(fuse, "lowFuse", info.lowFuseDescription);
}
// ----------------------------------------------------------------------------
// Convert the fuse extendted bits into readable string
// ----------------------------------------------------------------------------
QString
LarvorDevices::fuseExtendedToString(DeviceInfo const& info,
                                            unsigned char fuse)
{
    return fuseToString(fuse, "extendedFuse", info.extendedFuseDescription);
}

// ----------------------------------------------------------------------------
// Convert the locks into readable string
// ----------------------------------------------------------------------------
QString 
LarvorDevices::lockToString(unsigned char lock)
{
    QString txt ="";
    txt += "LB1: ";
    if (lock & (1 << 0)) {
        txt += "1";
    } else {
        txt += "0";
    }
    txt += "\nLB2: ";
    if (lock & (1 << 1)) {
        txt += "1";
    } else {
        txt += "0";
    }
    txt += "\nBLB01: ";
    if (lock & (1 << 2)) {
        txt += "1";
    } else {
        txt += "0";
    }
    txt += "\nBLB02: ";
    if (lock & (1 << 3)) {
        txt += "1";
    } else {
        txt += "0";
    }
    txt += "\nBLB11: ";
    if (lock & (1 << 4)) {
        txt += "1";
    } else {
        txt += "0";
    }
    txt += "\nBLB12: ";
    if (lock & (1 << 5)) {
        txt += "1";
    } else {
        txt += "0";
    }
    return txt;
}
