/****************************************************************************
**
** Copyright (C) 2010 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    mainwindow.cpp
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @brief   Interface between serial port and JTAG/SPI for programming AVR microcontrollers
// @web     http://sophiateam.undrgnd.free.fr
// @date    08/08/2010
// ---------------------------------------------------------------------------

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "writeFuse.h"
#include <QMessageBox>
#include <QTime>
#include <QFileDialog>
#include <QProgressBar>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent),
    settings_("LsmSoftware", "larvor"),
    progress_(new QProgressBar(this)),
    ui(new Ui::MainWindow),
    connected_(false),
    larve_(&serialPort_)
{
    int devicesCount = LarvorDevices::loadDeviceConfigFile("larvorDevices.xml");
	
    loadDirName();
    comPort_ = settings_.value("Port", "COM1").toString();

    ui->setupUi(this);
    updateCOMList();
    progress_->setVisible(false);
    statusBar()->addWidget(progress_);

    connect(larve_.getProgress(), SIGNAL(valueChanged(double)),
            this, SLOT(progressChanged(double)));
    connect(progress_, SIGNAL(canceled ()),
            larve_.getProgress(), SLOT(setCanceled()));

    larve_.setProgrammingMode((Larvor::ProgrammingMode)settings_.value("ProgrammingMode", Larvor::jtagProgrammingMode).toInt());
    ui->spiRB->setChecked(larve_.getProgrammingMode() == Larvor::spiProgrammingMode);
    ui->jtagRB->setChecked(larve_.getProgrammingMode() == Larvor::jtagProgrammingMode);

    QFont f("Courier");
    ui->comTextEdit->setFont(f);

    if (devicesCount > 0)
        ui->comTextEdit->append("The configuration file 'larvorDevices.xml' has been loaded. " + QString::number(devicesCount) + " devices found");
    else
        ui->comTextEdit->append("The configuration file 'larvorDevices.xml' was not found or it does not contain any device");
}

MainWindow::~MainWindow()
{
    settings_.setValue("Port", comPort_);
    delete ui;
}

// ===========================================================================
// Serial port management
// ===========================================================================
void MainWindow::updateCOMList()
{
    if (connected_) {
        return;
    }
    ui->portCB->clear();
    for(int i=1; i < 10; ++i) {
        QString port = "COM" + QString::number(i);
        if (sirSerial::exists((const char*)port.toAscii())) {
            ui->portCB->addItem(port);
        }
    }
    int index = ui->portCB->findText(comPort_);
    if (index<0) index=0;
    if (ui->portCB->count() == 0) {
        ui->portCB->setEnabled(false);
    } else {
        ui->portCB->setEnabled(true);
        ui->portCB->setCurrentIndex (index);
        comPort_ = ui->portCB->itemText(index);
    }
}

void MainWindow::on_portCB_currentIndexChanged(QString itemText)
{
    comPort_ = itemText;
}

void MainWindow::connectSerial()
{
    bool result = serialPort_.open((const char*)comPort_.toAscii(), 38400);
    if (result)
        connected();
    else
        cannotConnect("Cannot open serial port");
}
void MainWindow::disconnectSerial()
{
    serialPort_.close();
    disconnected();
}


void MainWindow::connected()
{
    connected_ = true;
    ui->portCB->setEnabled(!connected_);
    ui->SerialPortLabel->setEnabled(!connected_);
}

void MainWindow::cannotConnect(QString message)
{
    connected_ = false;
    ui->portCB->setEnabled(!connected_);
    ui->SerialPortLabel->setEnabled(!connected_);
    QMessageBox::information(this,
                             tr("Serial port Error"),
                             message,
                             QMessageBox::Ok);
}

void MainWindow::disconnected()
{
    connected_ = false;
    ui->portCB->setEnabled(!connected_);
    ui->SerialPortLabel->setEnabled(!connected_);
}

// ===========================================================================
// Serial port data parsing
// ===========================================================================
void MainWindow::dataReceived()
{
}

// ===========================================================================
// Sending data on the serial port
// ===========================================================================
void MainWindow::writeCOM(QList<unsigned char> const& buffer)
{
    QString txt = "> ";
    QListIterator<unsigned char> it(buffer);
    while(it.hasNext()) {
        unsigned char v = it.next();
        if (v < 16) {
            txt += "0x0"+QString::number(v, 16)+" ";
        } else {
            txt += "0x"+QString::number(v, 16)+" ";
        }
        serialPort_.write(v);
    }
    ui->comTextEdit->append(txt);
}

void MainWindow::writeCOM(unsigned char data)
{
    QList<unsigned char> buffer;
    buffer.push_back(data);
    writeCOM(buffer);
}


// ===========================================================================
// Commands
// ===========================================================================

void MainWindow::on_actionExit_triggered()
{
    this->close();
}

void MainWindow::on_actionAbout_triggered()
{
    QString title = "About Larvor";
    QString message = "<html><body>";
    message += "Larvor v1.0 08/08/2010<br>\n";
    message += "Author: Laurent Saint-Marcel<br>\n";
    message += "<a href=\"http://sophiateam.undrgnd.free.fr\">http://sophiateam.undrgnd.free.fr</a><br>\n";
    message += "<br>\n";
    message += "Programation board for AVR throught the JTAG interface";
    message += "</body></html>";
    QMessageBox::about(this, title, message);
}

// ===========================================================================
// Commands
// ===========================================================================
MainWindow::CommandGuard::CommandGuard(MainWindow* process, QString const& commandName) :
        process_(process),
        commandName_(commandName),
        status_(false)
{
    chrono_.start();
    process_->setProgressGlobalRange(100);
}
MainWindow::CommandGuard::~CommandGuard()
{
    process_->setProgressLocalRangeMax(101);
    int t = chrono_.elapsed();
    if (status_) {
        process_->ui->comTextEdit->setTextColor(Qt::darkGreen);
        process_->ui->comTextEdit->append("==> " + commandName_ + " successful in " + QString::number(t/1000., 'g', 2) + " seconds");
    } else {
        process_->ui->comTextEdit->setTextColor(Qt::darkRed);
        process_->ui->comTextEdit->append("==> " + commandName_ + " FAILED in " + QString::number(t/1000., 'g', 2) + " seconds");
    }
    process_->disconnectSerial();
}
void MainWindow::CommandGuard::setStatusSuccessful()
{
    status_ = true;
}
void MainWindow::CommandGuard::appendInfo(QString info)
{
    process_->ui->comTextEdit->setTextColor(Qt::black);
    process_->ui->comTextEdit->append("  " + info);
}
bool MainWindow::CommandGuard::testSerialPort()
{
    process_->connectSerial();
    process_->ui->comTextEdit->setTextColor(Qt::black);
    process_->ui->comTextEdit->append( "---------------------------------------------------------------------");
    process_->ui->comTextEdit->setTextColor(Qt::darkBlue);
    process_->ui->comTextEdit->append("> " + commandName_ + ":");
    if (!process_->connected_ || !process_->larve_.ping()) {
        process_->ui->comTextEdit->setTextColor(Qt::darkRed);
        appendInfo("Larvor board NOT detected. Check the serial port!");
        return false;
    } else {
        return true;
    }
}

// ----------------------------------------------------------------------------
// testSerialPort
// ----------------------------------------------------------------------------
void MainWindow::on_actionTestSerialPort_triggered()
{
    CommandGuard guard(this, "testSerialPort");
    if (guard.testSerialPort()) {
        guard.appendInfo("Larvor board detected!");
        guard.setStatusSuccessful();
    }
}

// ----------------------------------------------------------------------------
// getDeviceId
// ----------------------------------------------------------------------------
void MainWindow::on_IDCodeBtn_clicked()
{
    CommandGuard guard(this, "getDeviceId");

    int     manufacturerId;
    int     deviceId;
    LarvorDevices::DeviceInfo info;
    if (larve_.getDeviceId(manufacturerId,  deviceId)
        && LarvorDevices::findDeviceInfo(manufacturerId,  deviceId, info))
    {
        guard.appendInfo("Detected: " + info.deviceName + " " + info.manufacturerName);
        guard.setStatusSuccessful();
    }
}

// ----------------------------------------------------------------------------
// getDeviceInfo
// ----------------------------------------------------------------------------
bool MainWindow::getDeviceInfo(LarvorDevices::DeviceInfo& info, bool dumpInfo)
{
    if (connected_) {
        int manufacturerId;
        int deviceId;
        if (larve_.getDeviceId(manufacturerId,  deviceId)) {
            if (LarvorDevices::findDeviceInfo(manufacturerId,  deviceId, info)) {
                ui->comTextEdit->setTextColor(Qt::black);
                ui->comTextEdit->append("  Detected: " + info.deviceName + " " + info.manufacturerName);
                if (dumpInfo) {
                    ui->comTextEdit->append("    EEprom: Size=" + QString::number(info.eepromSize) +" pageSize=" + QString::number(info.eepromPageSize));
                    ui->comTextEdit->append("    Flash: Size=" + QString::number(info.flashSize) +" pageSize=" + QString::number(info.flashPageSize));
                }
                return true;
            }
        }
    }
    ui->comTextEdit->setTextColor(Qt::darkRed);
    ui->comTextEdit->append("Device not detected");
    return false;
}

// ----------------------------------------------------------------------------
// getSignature
// ----------------------------------------------------------------------------
void MainWindow::on_actionGetSignature_triggered()
{
    CommandGuard guard(this, "getSignature");
    if (guard.testSerialPort()) {
        LarvorDevices::DeviceInfo info;
        getDeviceInfo(info, true);
        unsigned char signature1, signature2, signature3;
        if (larve_.readSignature(signature1, signature2, signature3)) {
          //  guard.appendInfo("Signature: " + LarvorDevices::signatureToString(signature1, signature2, signature3) );
            guard.setStatusSuccessful();
        }
    }
}

// ----------------------------------------------------------------------------
// readFuse
// ----------------------------------------------------------------------------
void MainWindow::on_actionReadFuse_triggered()
{
    CommandGuard guard(this, "readFuse");
    if (guard.testSerialPort()) {
	
		LarvorDevices::DeviceInfo info;
		if (!getDeviceInfo(info))
			return;
			
        unsigned char fuseLow, fuseHigh, fuseExtended;
        if (larve_.readFuseHigh(fuseHigh)
            && larve_.readFuseLow(fuseLow)
            && larve_.readFuseExt(fuseExtended))
        {
            guard.appendInfo("Fuse:\n"
                             + LarvorDevices::fuseLowToString(info, fuseLow)
                             + "\n"
                             + LarvorDevices::fuseHighToString(info, fuseHigh)
                             + "\n"
                             + LarvorDevices::fuseExtendedToString(info, fuseExtended) );
            guard.setStatusSuccessful();
        }
    }
}

// ----------------------------------------------------------------------------
// writeFuse
// ----------------------------------------------------------------------------
void MainWindow::on_actionWriteFuse_triggered()
{
    CommandGuard guard(this, "writeFuse");
    if (guard.testSerialPort()) {
	
		LarvorDevices::DeviceInfo info;
		if (!getDeviceInfo(info))
			return;
		
        unsigned char fuseLow=0xE2, fuseHigh=0xFF, fuseExtended=0xFF;
        if ((larve_.readFuseHigh(fuseHigh)
            && larve_.readFuseLow(fuseLow)
            && larve_.readFuseExt(fuseExtended)))
            {
            WriteFuseDialog dialog(this, info, fuseHigh, fuseLow, fuseExtended);
            if (dialog.exec() == QDialog::Accepted) {
                bool result = false;
                switch(dialog.getFuseMode()) {
                case WriteFuseDialog::highFuseMode:
                    result = larve_.writeFuseHigh(dialog.getFuseHigh());
                    result &= larve_.readFuseHigh(fuseHigh);
                    result &= (fuseHigh == dialog.getFuseHigh());
                    break;
                case WriteFuseDialog::lowFuseMode:
                    result = larve_.writeFuseLow(dialog.getFuseLow());
                    result &= larve_.readFuseLow(fuseLow);
                    result &= (fuseLow == dialog.getFuseLow());
                    break;
                case WriteFuseDialog::extendedFuseMode:
                    result = larve_.writeFuseExt(dialog.getFuseExtended());
                    result &= larve_.readFuseExt(fuseExtended);
                    result &= (fuseExtended == dialog.getFuseExtended());
                    break;
                }
            }
            guard.setStatusSuccessful();
        }
    }
}

// ----------------------------------------------------------------------------
// readEEprom
// ----------------------------------------------------------------------------
void MainWindow::on_actionReadEEprom_triggered()
{
    QString fileName = QFileDialog::getSaveFileName(this,
                                                    tr("Select EEprom file"), 
                                                    dirname_+ "/eeprom.hex",
                                                    tr("Hex Files (*.hex)"));
    updateDirName(fileName);

    CommandGuard guard(this, "readEEprom");
    if (!guard.testSerialPort()) {
        return;
    }

    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    QMap<unsigned long, unsigned char> eeprom;
    setProgressLocalRangeMax(100);
    if (larve_.readEEprom(0, info.eepromSize, eeprom)) {
        guard.setStatusSuccessful();
    }

    LarvorFiles::writeEepromIntelHex(fileName, eeprom);
    guard.appendInfo("EEprom saved in " + fileName);
}

// ----------------------------------------------------------------------------
// writeEEprom
// ----------------------------------------------------------------------------
void MainWindow::on_actionWriteEEprom_triggered()
{
    QMap<unsigned long, unsigned char> eeprom;
    QMap<unsigned long, unsigned char>::const_iterator it;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select EEprom file"), 
                                                    dirname_,
                                                    tr("Hex Files (*.hex)"));
    updateDirName(fileName);

    CommandGuard guard(this, "writeEEprom");
    if (!guard.testSerialPort()) {
        return;
    }
    setProgressLocalRangeMax(5);
    
    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    if (!LarvorFiles::readEepromFile(fileName, eeprom) || eeprom.empty()) {
        guard.appendInfo("ERROR: " + fileName + " is corrupted");
    }

    unsigned long addrBegin=0, addrEnd=0;
    unsigned char pagesize=info.eepromPageSize;
    LarvorFiles::adjustToPageSize(eeprom, pagesize, addrBegin, addrEnd);

    if (addrBegin > addrEnd || addrEnd > info.eepromSize) {
        guard.appendInfo("ERROR: There is not enought space in the device to write "
                         + fileName
                         + ". File data size=0x" + QString::number(addrEnd, 16)
                         + ", Flash size=0x" + QString::number(info.eepromSize, 16));
        return;
    }

    setProgressLocalRangeMax(100);
    if (larve_.writeEEprom(addrBegin, addrEnd, eeprom, pagesize)) {
        guard.appendInfo(fileName + " written in EEprom");
        guard.setStatusSuccessful();
    }
}

// ----------------------------------------------------------------------------
// checkEEprom
// ----------------------------------------------------------------------------
void MainWindow::on_actionCheckEEprom_triggered()
{
    QMap<unsigned long, unsigned char> eeprom;
    QMap<unsigned long, unsigned char> eepromCheck;
    QMap<unsigned long, unsigned char>::const_iterator it;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select EEprom file"),
                                                    dirname_,
                                                    tr("Hex Files (*.hex)"));
    updateDirName(fileName);

    CommandGuard guard(this, "checkEEprom");
    if (!guard.testSerialPort()) {
        return;
    }

    setProgressLocalRangeMax(2);
    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    if (!LarvorFiles::readEepromFile(fileName, eeprom) || eeprom.empty()) {
        guard.appendInfo("ERROR: " + fileName + " is corrupted");
    }

    unsigned long addrBegin=0, addrEnd=0;
    unsigned char pagesize=info.eepromPageSize;
    LarvorFiles::adjustToPageSize(eeprom, pagesize, addrBegin, addrEnd);
    if (addrBegin > addrEnd || addrEnd > info.flashSize) {
        guard.appendInfo("ERROR: There is not enought space in the device to write "
                         + fileName
                         + ". File data size=0x" + QString::number(addrEnd, 16)
                         + ", Flash size=0x" + QString::number(info.eepromSize, 16));
        return;
    }

    ui->comTextEdit->repaint();
    setProgressLocalRangeMax(99);

    if (larve_.readEEprom(addrBegin, addrEnd, eepromCheck)
        && LarvorFiles::checkEEprom(eeprom, eepromCheck))
    {
        guard.setStatusSuccessful();
    } else {
        guard.appendInfo("ERROR: EEprom and file do not match!");
    }
}

// ----------------------------------------------------------------------------
// readFlash
// ----------------------------------------------------------------------------
void MainWindow::on_actionReadFlash_triggered()
{
    QString fileName = QFileDialog::getSaveFileName(this,
                                                    tr("Select Flash file"), 
                                                    dirname_ + "/flash.rom",
                                                    tr("Rom Files (*.rom)"));
    updateDirName(fileName);

    CommandGuard guard(this, "readFlash");
    if (!guard.testSerialPort()) {
        return;
    }

    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    QMap<unsigned long, unsigned short> flash;
    setProgressLocalRangeMax(100);
    if (larve_.readFlash(0, info.flashSize, flash)) {
        guard.setStatusSuccessful();
    }

    LarvorFiles::writeFlashIntelHex(fileName, flash);
    guard.appendInfo("Flash saved in " + fileName);
}

// ----------------------------------------------------------------------------
// writeFlash
// ----------------------------------------------------------------------------
void MainWindow::on_actionWriteFlash_triggered()
{
    
    QMap<unsigned long, unsigned short> flash;
    QMap<unsigned long, unsigned char>::const_iterator it;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select Flash file"), 
                                                    dirname_,
                                                    tr("Rom Files (*.rom)"));
    updateDirName(fileName);
    
    CommandGuard guard(this, "writeFlash");
    if (!guard.testSerialPort()) {
        return;
    }
    setProgressLocalRangeMax(2);
    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    if (!LarvorFiles::readFlashFile(fileName, flash)) {
        guard.appendInfo("ERROR: " + fileName + " is corrupted");
        return;
    }
    unsigned long addrBegin=0, addrEnd=0;
    unsigned char pagesize=info.flashPageSize;
    LarvorFiles::adjustToPageSize(flash, pagesize, addrBegin, addrEnd);
    if (addrBegin > addrEnd || addrEnd > info.flashSize) {
        guard.appendInfo("ERROR: There is not enought space in the device to write "
                         + fileName
                         + ". File data size=0x" + QString::number(addrEnd, 16)
                         + ", Flash size=0x" + QString::number(info.flashSize, 16));
        return;
    }

    int t;
    QTime chrono2;
    chrono2.start();
    guard.appendInfo("Chip Erasing..."); ui->comTextEdit->repaint();
    setProgressLocalRangeMax(5);
    if (!larve_.chipErase()) {
        guard.appendInfo("Chip NOT erased\n");
        return;
    }
    t=chrono2.elapsed(); 
    guard.appendInfo("Chip erased in " + QString::number(t/1000., 'g', 2) + " seconds");

    chrono2.restart();
    guard.appendInfo("Writing flash..."); ui->comTextEdit->repaint();
    setProgressLocalRangeMax(100);
    if (larve_.writeFlash(addrBegin, addrEnd, flash, pagesize)){
        t=chrono2.elapsed();
        guard.appendInfo("Flash written and verified in " + QString::number(t/1000., 'g', 2) + " seconds");
        guard.setStatusSuccessful();
    } 

}

// ----------------------------------------------------------------------------
// checkFlash
// ----------------------------------------------------------------------------
void MainWindow::on_actionCheckFlash_triggered()
{
    QMap<unsigned long, unsigned short> flash;
    QMap<unsigned long, unsigned short> flashCheck;
    QMap<unsigned long, unsigned char>::const_iterator it;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select Flash file"), 
                                                    dirname_,
                                                    tr("Rom Files (*.rom)"));
    updateDirName(fileName);
    
    CommandGuard guard(this, "checkFlash");
    if (!guard.testSerialPort()) {
        return;
    }

    setProgressLocalRangeMax(2);
    LarvorDevices::DeviceInfo info;
    if (!getDeviceInfo(info))
        return;

    if (!LarvorFiles::readFlashFile(fileName, flash)) {
        guard.appendInfo("ERROR: " + fileName + " is corrupted");
        return;
    }

    unsigned long addrBegin=0, addrEnd=0;
    unsigned char pagesize=info.flashPageSize;
    LarvorFiles::adjustToPageSize(flash, pagesize, addrBegin, addrEnd);
    if (addrBegin > addrEnd || addrEnd > info.flashSize) {
        guard.appendInfo("ERROR: There is not enought space in the device to write "
                         + fileName
                         + ". File data size=0x" + QString::number(addrEnd, 16)
                         + ", Flash size=0x" + QString::number(info.flashSize, 16));
        return;
    }

    ui->comTextEdit->repaint();
    setProgressLocalRangeMax(99);
    if (larve_.readFlash(addrBegin, addrEnd, flashCheck)
        && LarvorFiles::checkFlash(flash, flashCheck))
        {
        guard.setStatusSuccessful();
    } else {
        guard.appendInfo("ERROR: Flash and file do not match!");
    } 
}

// ----------------------------------------------------------------------------
// chipErase
// ----------------------------------------------------------------------------
void MainWindow::on_actionEraseChip_triggered()
{
    CommandGuard guard(this, "chipErase");
    if (guard.testSerialPort()) {
        if (larve_.chipErase()) {
            guard.appendInfo("Chip erased");
            guard.setStatusSuccessful();
        } else {
            guard.appendInfo("Chip NOT erased\n");
        }
    }
}

// ============================================================================
// Progress bar
// ============================================================================
void MainWindow::setProgressLocalRangeMax(int max)
{
    progressMin_ = progressMax_;
    progressMax_ = max;
    if (max > progress_->maximum()) {
        progress_->setVisible(false);
    } else {
        progress_->setValue(progressMin_);
    }
}

void MainWindow::setProgressGlobalRange(int max)
{
    progressMin_ = 0;
    progress_->setRange(0, max);
    progress_->setValue(0);
    progress_->setVisible(true);
    larve_.getProgress()->setCanceled(false);
}

void MainWindow::progressChanged(double v)
{
    // callback when the serial port communicates
    progress_->setValue(progressMin_ + (int)(v*(progressMax_-progressMin_)));
}

// ============================================================================
// PLastDirName
// ============================================================================
void MainWindow::updateDirName(QString const& filename)
{
    QFileInfo fileInfo(filename);
    dirname_ =  fileInfo.absolutePath();
    settings_.setValue("LastDirName", dirname_);
}
void MainWindow::loadDirName()
{
    dirname_ = settings_.value("LastDirName").toString();
}

void MainWindow::on_jtagRB_clicked()
{
    larve_.setProgrammingMode(Larvor::jtagProgrammingMode);
    settings_.setValue("ProgrammingMode", larve_.getProgrammingMode());
}

void MainWindow::on_spiRB_clicked()
{
    larve_.setProgrammingMode(Larvor::spiProgrammingMode);
    settings_.setValue("ProgrammingMode", larve_.getProgrammingMode());
}

