/* SP12: A serial programmer for working with Atmel AVR uCs          */
/* Copyright (C) 1997-2003 Ken Huntington, Kevin Towers, Pitronics.  */

/* This program is free software; you can redistribute it and/or     */
/* modify it under the terms of the GNU General Public License       */
/* as published by the Free Software Foundation; either version 2    */
/* of the License, or (at your option) any later version.            */

/* This program is distributed in the hope that it will be useful,   */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
/* GNU General Public License for more details.                      */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software       */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,            */
/* MA  02111-1307, USA.                                              */

/* Pitronics can be reached by email: sbolt@xs4all.nl                */
/* Kevin Towers can be reached by email: ktowers@omnexcontrols.com   */
/* Ken Huntington can be reached by email: kenh@compmore.net         */

#include "Larvor.h"
#include <stdio.h>


namespace LarvorFiles
{

#define WORDLEN                 16
#define DATALEN                 5
#define FLASH_DATAPTR           6
#define MAXLEN                  500
#define BYTES_PER_LINE          16

// ----------------------------------------------------------------------------
// Routines to write the eeprom buffer into a file,        
// using Intel HEX format or as a hex dump with ascii translation.   
// ----------------------------------------------------------------------------
bool writeEepromIntelHex (QString const& filename, 
                          QMap<unsigned long, unsigned char> const& buffer) 
{
    FILE *filePtr = fopen((const char*)filename.toAscii(), "w");
    if (filePtr == NULL)
        return false;

    long length = buffer.size();
    QMapIterator<unsigned long, unsigned char> it(buffer);
    int line_cnt = 0;                     // line byte count         
    int line_len = 0;                     // expected line length     
    unsigned char checksum = 0;
    while (length > 0 && it.hasNext()) {
        it.next();

        unsigned long addr = it.key();
        unsigned char data = it.value();

        if (line_cnt == 0) {               // write line header        
            line_len = (length > BYTES_PER_LINE) ? BYTES_PER_LINE : length;
            fprintf (filePtr, ":%02X%04X00", line_len, (unsigned int)addr);
            checksum = line_len;
            checksum += addr & 0xFF;
            checksum += addr >> 8;
        }
        fprintf (filePtr, "%02X", data);
        checksum += data & 0xFF;
        if (++line_cnt == line_len) {      // write checksum           
            fprintf (filePtr, "%02X\n", -checksum & 0xFF);
         line_cnt = 0;
        }
        length--;
    }
    fprintf (filePtr, ":00000001FF\n");   // write the terminator   
    fclose(filePtr);
    return true;
}

// ----------------------------------------------------------------------------
// Routines to write the flash buffer into a file,        
// using Intel HEX format or as a hex dump with ascii translation.   
// ----------------------------------------------------------------------------

bool writeFlashIntelHex (QString const& filename, 
                         QMap<unsigned long, unsigned short> const& buffer) 
{

    FILE *filePtr = fopen((const char*)filename.toAscii(), "w");
    if (filePtr == NULL)
        return false;

    long length = buffer.size();
    QMapIterator<unsigned long, unsigned short> it(buffer);
    
    int line_cnt = 0;                     // line byte count          
    int line_len = 0;                     // expected line length     
    unsigned char checksum = 0;
    unsigned long  addr=0;
    unsigned short data=0;

    if (it.hasNext()) {
        it.next();
        addr = it.key();
        data = it.value();
    }

    length *= 2;
    while (length > 0) {
        if (addr > 0xFFFF) {               // add record type 02       
            fprintf (filePtr, ":020000021000EC\n");
            addr = 0x10000 - addr;
        }
        if (line_cnt == 0) {               // write line header        
            line_len = (length > BYTES_PER_LINE) ? BYTES_PER_LINE : length;
            fprintf (filePtr, ":%02X%04lX00", line_len, addr);
            checksum = line_len;
            checksum += addr & 0xFF;
            checksum += addr >> 8;
        }
        
        unsigned char byt;
        if (addr & 0x0001) {               // is it an odd address?    
            byt = data >> 8;             // write MSB 
            if (it.hasNext()) {
                it.next();
                data = it.value();
            }
            //addr = it.key();
        } else {
            byt = data & 0x00FF;         // write LSB                
        }
        fprintf (filePtr, "%02X", byt);
        checksum += byt;
        
        if (++line_cnt == line_len) { // write checksum                
            fprintf (filePtr, "%02X\n", -checksum & 0xFF);
            line_cnt = 0;
        }
        addr++;
        length--;
    }
    fprintf (filePtr, ":00000001FF\n");   // write the terminator     
    fclose(filePtr);
    return true;
}

/* Reads program file in Atmel generic format (address:data) or      */
/* Intel HEX format; stores data into into flashBuf by address.      */
/* flashBuf by address. The two numbers in the generic format are    */
/* hex without `0x'. For example: 00000c:99e1                        */
/* Note: the 16-bit bufLimit is not suitable for Mega address space. */

bool readFlashFile(QString const& filename, 
                   QMap<unsigned long, unsigned short> & flashBuf) 
{

    FILE *flashBufPtr;
    char wordstring[WORDLEN];
    unsigned long address;
    unsigned int data;
    char buffer[MAXLEN];
    long offsetAddress = 0L;
    int dataByteCount;
    unsigned int lineAddress;
    int recordType;
    int idx;
    int checksum;
    int tmp;
    
    if ((flashBufPtr = fopen((const char*)filename.toAscii(), "r")) == NULL) {
        return false;
    } 

    if (fgets(wordstring, WORDLEN, flashBufPtr) != NULL) {
        /*
         * Try Atmel generic format (default of the Atmel assembler).
         */ 
        if (wordstring[FLASH_DATAPTR] == ':') {
            fseek(flashBufPtr, 0, SEEK_SET);
            while (fgets(wordstring, WORDLEN, flashBufPtr) != NULL) {
                sscanf(wordstring, "%lx", &address);
                if (wordstring[FLASH_DATAPTR] != ':') {
                    return false;
                }
                sscanf(&wordstring[FLASH_DATAPTR + 1], "%x", &data);
                flashBuf[address] = data;
            }
            fclose(flashBufPtr);
            return true;
        }
        /*
         * Try Intel HEX format.
         */
        if (wordstring[0] == ':') {
            fseek(flashBufPtr, 0, SEEK_SET);
            while (fgets(buffer, MAXLEN, flashBufPtr) != NULL) {
                checksum = 0;
                sscanf(&buffer[1], "%2x", &dataByteCount);
                checksum += dataByteCount;
                sscanf(&buffer[3], "%4x", &lineAddress);
                sscanf(&buffer[3], "%2x", &tmp);
                checksum += tmp;
                sscanf(&buffer[5], "%2x", &tmp);
                checksum += tmp;
                sscanf(&buffer[7], "%2x", &recordType);
                checksum += recordType;
                /*
                 * Record type 02 means update offset address
                 */
                if (recordType == 2) {
                    sscanf(&buffer[9], "%4lx", &offsetAddress);
                    sscanf(&buffer[9], "%2x", &tmp);
                    checksum += tmp;
                    sscanf(&buffer[11], "%2x", &tmp);
                    checksum += tmp;
                    checksum = (unsigned char) (0x100 - (unsigned char) checksum);
                    sscanf(&buffer[13], "%2x", &tmp);
                    if (checksum != tmp) {
                        fclose(flashBufPtr);
                        return false;
                    }
                }
                if (recordType == 0) {
                    for (idx = 9; idx < 9 + dataByteCount * 2; idx += 2) {
                        sscanf(&buffer[idx], "%2x", &data);
                        /*
                         * Use word addresses for the AT90 2-byte data format
                         */
                        address = lineAddress / 2;
                        address =  address + (offsetAddress * 0x10) / 2;
                        if (lineAddress & 0x0001)   /* is it an odd address? */
                            /* then set MSB of data  */
                            flashBuf[address] = ((data<<8) & 0xff00) | (flashBuf[address] & 0x00ff);
                        else                       /* else set LSB of data  */
                            flashBuf[address] = (data & 0xff) | (flashBuf[address] & 0xff00);
                        lineAddress++;
                        checksum += data;
                    }                     
                    checksum = (unsigned char) (0x100 - (unsigned char) checksum);
                    sscanf(&buffer[idx], "%2x", &tmp);
                    if (checksum != tmp) {
                        fclose(flashBufPtr);
                        return false;
                    }
                }
            }
            fclose(flashBufPtr);
            return true;
        }
    }
    fclose(flashBufPtr);
    return false;
}

#if 0
/* Reads eeprom file in format address:data, stores data into         */
/* eepromBuf by address. Both numbers can be hexadecimal (0xC4),      */
/* decimal (196), octal (0304) or binary (B11000100).                 */
bool readEepromFile(QString const& filename,
                    QMap<unsigned long, unsigned char> & eepromBuf)
{
    FILE *eepromBufPtr;
    if ((eepromBufPtr = fopen((const char*)filename.toAscii(), "r")) == NULL) {
        return false;
    }
    int i=0, c;
    int addr=0;
    char buf[5];
    while(c = fgetc(eepromBufPtr) != EOF) {
        buf[i++] =c;
        if (i == 2) {
            i = 0;
            unsigned int data;
            sscanf(buf, "%2x", & data);
            eepromBuf[addr++] = data&0xFF;
        }
    }
    fclose(eepromBufPtr);
    return true;
}
#else
bool readEepromFile(QString const& filename, 
                    QMap<unsigned long, unsigned char> & eepromBuf) 
{
    
    FILE *eepromBufPtr;
    char wordstring[WORDLEN];
    unsigned long address;
    unsigned int data;
    char buffer[MAXLEN];
    unsigned long offsetAddress = 0L;
    int dataByteCount;
    unsigned int lineAddress;
    int recordType;
    int idx;
    int checksum;
    int tmp;
    
    if ((eepromBufPtr = fopen((const char*)filename.toAscii(), "r")) == NULL) {
        return false;
    } 
    if (fgets(wordstring, WORDLEN, eepromBufPtr) != NULL) {
        /*
         * Try Intel HEX format.
         */
        if (strchr(wordstring, ':') == &wordstring[0]) {
            fseek(eepromBufPtr, 0, SEEK_SET);
            while (fgets(buffer, MAXLEN, eepromBufPtr) != NULL) {
                checksum = 0;
                sscanf(&buffer[1], "%2x", &dataByteCount);
                checksum += dataByteCount;
                sscanf(&buffer[3], "%4x", &lineAddress);
                sscanf(&buffer[3], "%2x", &tmp);
                checksum += tmp;
                sscanf(&buffer[5], "%2x", &tmp);
                checksum += tmp;
                sscanf(&buffer[7], "%2x", &recordType);
                checksum += recordType;
                /*
                 * Record type 02 means update offset address
                 */
                if (recordType == 2) {
                    sscanf(&buffer[9], "%4lx", &offsetAddress);
                    sscanf(&buffer[9], "%2x", &tmp);
                    checksum += tmp;
                    sscanf(&buffer[11], "%2x", &tmp);
                    checksum += tmp;
                    checksum = (unsigned char) (0x100 - (unsigned char) checksum);
                    sscanf(&buffer[13], "%2x", &tmp);
                    if (checksum != tmp) {
                        fclose(eepromBufPtr);
                        return false;
                    }
                }
                if (recordType == 0) {
                    address = lineAddress;
                    address =  address + (offsetAddress * 0x10) / 2;
                    /*
                     * Use byte addresses for single byte eeprom data format.
                     */
                    for (idx = 9; idx < 9 + dataByteCount * 2; idx += 2) {
                        sscanf(&buffer[idx], "%2x", &data);
                        eepromBuf[address] = data;
                        address++;
                        checksum += data;
                    }
                    checksum = (unsigned char) (0x100 - (unsigned char) checksum);
                    sscanf(&buffer[idx], "%2x", &tmp);
                    if (checksum != tmp) {
                        fclose(eepromBufPtr);
                        return false;
                    }
                }
            }
            fclose(eepromBufPtr);
            return true;
         }
    }
    return false;
}
#endif

    void adjustToPageSize(QMap<unsigned long, unsigned char>& data,
                          unsigned int pagesize,
                          unsigned long &addrBegin,
                          unsigned long &addrEnd)
    {
        QMap<unsigned long, unsigned char >::const_iterator it;

        it = data.begin();
        addrBegin = it.key();
        while ((addrBegin % pagesize) != 0) {
            --addrBegin;
        }

        it = data.end(); --it;
        addrEnd = it.key();
        while ((addrEnd % pagesize) != (pagesize-1)) {
            ++addrEnd;
        }

        unsigned long addr;
        for(addr=addrBegin; addr <= addrEnd; ++addr) {
            if (data.find(addr) == data.end())
                data[addr] = 0xFF;
        }

    }

    void adjustToPageSize(QMap<unsigned long, unsigned short>& data,
                          unsigned int pagesize,
                          unsigned long &addrBegin,
                          unsigned long &addrEnd)
    {
        QMap<unsigned long, unsigned short>::const_iterator it;

        it = data.begin();
        addrBegin = it.key();
        while ((addrBegin % pagesize) != 0) {
            --addrBegin;
        }

        it = data.end(); --it;
        addrEnd = it.key();
        while ((addrEnd % pagesize) != (pagesize-1)) {
            ++addrEnd;
        }

        unsigned long addr;
        for(addr=addrBegin; addr <= addrEnd; ++addr) {
            if (data.find(addr) == data.end())
                data[addr] = 0xFFFF;
        }

    }

    bool checkFlash(QMap<unsigned long, unsigned short> const& data,
                    QMap<unsigned long, unsigned short> const& data2)
    {
        if (data.size() != data2.size()) {
            return false;
        }
        QMap<unsigned long, unsigned short>::const_iterator it, it2;
        for(it = data.begin(); it != data.end(); ++it) {
            it2 = data2.find(it.key());
            if (it2 == data2.end() || it2.value() != it.value()) {
                return false;
            }
        }
        return true;
    }

    bool checkEEprom(QMap<unsigned long, unsigned char> const& data,
                     QMap<unsigned long, unsigned char> const& data2)
    {
        if (data.size() != data2.size()) {
            return false;
        }
        QMap<unsigned long, unsigned char>::const_iterator it, it2;
        for(it = data.begin(); it != data.end(); ++it) {
            it2 = data2.find(it.key());
            if (it2 == data2.end() || it2.value() != it.value()) {
                return false;
            }
        }
        return true;
    }
}; // end of namespace
