/****************************************************************************
**
** Copyright (C) 2007 Laurent Saint-Marcel. All rights reserved.
**
** This file may be used under the terms of the GNU General Public
** License version 3.0 as published by the Free Software Foundation
** and appearing in the file gpl-3.0.txt included in the packaging of
** this file. 
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/
// ---------------------------------------------------------------------------
// @file    sirSerial.h
// @author  Laurent Saint-Marcel (lstmarcel@yahoo.fr)
// @web     http://craid.free.fr
// @date    08/05/2001
// ---------------------------------------------------------------------------
// communication on a serial port for windows
// ---------------------------------------------------------------------------

#ifndef __SIR_SERIAL_H__
#define __SIR_SERIAL_H__

#include <QThread>
#include <QMutex>

#ifdef WIN32
#include <windows.h>
#include <commctrl.h>
#else
#include <termios.h>
#define NOPARITY 0
#define ONESTOPBIT 1
#endif

#define DEFAULT_READ_TIMEOUT      500      // milliseconds

#ifdef SERIAL_ERROR
#undef SERIAL_ERROR
#endif

#ifdef SERIAL_SUCCESS
#undef SERIAL_SUCCESS
#endif


#define SERIAL_ERROR	-1
#define SERIAL_SUCCESS	0

#define DEFAULT_PORT_NAME	"COM1"

// ---------------------------------------------------------------------------
// class sirSerial
// ---------------------------------------------------------------------------
class sirSerial : public QThread {

  public :
         sirSerial();
         ~sirSerial();
    bool open(const char* gszPort = DEFAULT_PORT_NAME, 
             long speed = 9600, 
             int parity = NOPARITY, 
             int bitslength = 8, 
             int stopbit = ONESTOPBIT);
    
    bool close();
    const char* getErrorMessage() const { return errorMessage_; }
          
    bool read(unsigned char * buffer, 
              unsigned long & length);
    bool forceRead(unsigned char * lpBuf, 
                   unsigned long &length,
                   int retry=-2); // infinite retry
    
    bool write(unsigned char * buffer, 
               unsigned long length);
    bool write(unsigned char buffer);

    void usleep(unsigned long t) { QThread::usleep(t); };

    // return true if the serial port exists
    static bool exists(const char* gszPort);
    
  private :
#ifdef WIN32
    WCHAR gszPort_[256];
    HANDLE CommPortHandle; //Holds the handle to the communications port returned by windows.
    COMMTIMEOUTS CommPortTimeOut; // This data structure is programmed with timeout parameters for the serial port
    DCB CommPortDCB; // This data structure is programmed with the serial port parameters.
    short PortInitSuccessful; // This data member is assigned 1 if the serial port is successfully initialized, otherwise it is 0.
    unsigned long BytesTransferred; // Holds the number of bytes transferred during a serial port read or write operation. This is here just as a placeholder for windows API functions.
#else
    struct termios oldtio_;
    int fd_;
    int ttyNbr_;
#endif
    char errorMessage_[256];
};


class SerialPort : public QThread {
     Q_OBJECT

 public:
     SerialPort(QObject *parent = 0);
     virtual ~SerialPort();

     bool read(QList<unsigned char>& buffer);
     bool isConnected() const { return isConnected_; }

 signals:
     void connected();
     void cannotConnect(QString message);
     // Com port is closed. Signal sent by 'disconnect'
     void disconnected();
     void dataReceived(); 
     
 
 public slots:
     // Open the serial port to communication with the SI master Station
     // Call the following signals when finished: connected() or
     // cannotConnect()
     void open(QString const& portName, int baudRate);
	 
     bool write(QList<unsigned char> const& buffer);
     bool write(unsigned char data);
     // Close the serial port to communication
     // Call the following signals when finished: disconnected()
     void close();
     // Kill the communication thread - it can be restarted by creating
     // a new instance of QSerialPort
     void killThread();

 protected:
     enum ThreadRequest {
         sircNotConnected,
         sircToConnect,
         sircConnected,
         sircToDisconnect,
         sircToKill,

         sircWrite
     };

     void             run();
     virtual void     serialWrite();
     virtual void     serialRead();
     
     bool             setRequest(ThreadRequest r);
     ThreadRequest    getMode();

 protected:

     QMutex            mutex_;
     sirSerial*        serial_;
     ThreadRequest     request_;

     QString           portName_;
     int               baudRate_;
     bool              needHasDataSignal_;
     bool              isConnected_;

     QList<unsigned char> readBuffer_;
     QList<unsigned char> writeBuffer_;

};
#endif
